package multipart_form

  1. Overview
  2. Docs

Module Multipart_form.Content_typeSource

Sourcemodule Type : sig ... end
Sourcemodule Subtype : sig ... end
Sourcemodule Parameters : sig ... end
Sourcetype t = {
  1. ty : Type.t;
  2. subty : Subtype.t;
  3. parameters : (string * Parameters.value) list;
}
Sourceval equal : t -> t -> bool
Sourceval pp : t Fmt.t
Sourceval of_string : string -> (t, [> `Msg of string ]) result

of_string str returns the Content-Type value from a string which come from your HTTP stack. NOTE: the string must finish with "\r\n". If you are not sure about the value returned by your HTTP stack, you should append it.

Sourceval to_string : t -> string